
// -*- mode: c++; c-basic-offset:4 -*-

// This file is part of libnc-dods, A C++ implementation of netCDF 3 which
// supports reading from local files and OPeNDAP servers.

// Copyright (c) 2004 OPeNDAP, Inc.
// Author: James Gallagher <jgallagher@opendap.org>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 
#include <cppunit/TextTestRunner.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/extensions/HelperMacros.h>

// Tests for NCStructure. 02/24/04 jhrg

#include <algorithm>

#include <netcdf.h>

#include "NCConnect.h"
#include "NCSequence.h"
#include "NCStructure.h"
#include "NCArray.h"
#include "NCByte.h"
#include "nc_util.h"

#include "debug.h"

using namespace CppUnit;
using namespace std;

static void
delete_btp(BaseType *btp)
{
    delete btp; btp = 0;
}

class NCStructureTest : public TestFixture {
private:
    NCStructure *s;
    ClientParams *cp;
    NCTypeFactory *factory;
public: 
    NCStructureTest() {}
    ~NCStructureTest() {}

    void setUp() {
        cp = new ClientParams("http://localhost/");
        s = new NCStructure("test");
        factory = new NCTypeFactory;
    }

    void tearDown() {
        delete factory;
        delete s;
        delete cp;
    }
#if 0    
    void compare_list_to_vector(list<BaseType*> l, 
                                vector<BaseType*>::iterator vi,
                                vector<BaseType*>::iterator ve) {
        list<BaseType*>::iterator li = l.begin();
        while (li != l.end() && vi != ve) {
            DBG2((*vi)->print_decl(cerr));
            DBG2((*li)->print_decl(cerr));
            CPPUNIT_ASSERT((*li)->name() == (*vi)->name());
            CPPUNIT_ASSERT((*li)->type() == (*vi)->type());
            
            switch((*li)->type()) {
                case dods_sequence_c: {
                    NCSequence &ncs = dynamic_cast<NCSequence&>(**li);
                    VarList sv = ncs.get_variables_list();
    
                    compare_list_to_vector(sv, ncs.var_begin(), ncs.var_end());
                    break;
                }
                
                case dods_structure_c: {
                    NCStructure &ncs = dynamic_cast<NCStructure&>(**li);
                    VarList sv = ncs.get_variables_list();
    
                    compare_list_to_vector(sv, ncs.var_begin(), ncs.var_end());
                    break;
                }
                
                default:
                    break;
            }

            ++li; ++vi;
        }
    }
#endif
    CPPUNIT_TEST_SUITE( NCStructureTest );

    CPPUNIT_TEST(flatten_test1);
    CPPUNIT_TEST(flatten_test2);
    CPPUNIT_TEST(flatten_test3);
    CPPUNIT_TEST(flatten_test4);
    CPPUNIT_TEST(flatten_test5);
    CPPUNIT_TEST(flatten_test6);
    CPPUNIT_TEST(flatten_test7);

    CPPUNIT_TEST_SUITE_END();

    // A Structure with two atomic types
    void flatten_test1() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i32 = factory->NewInt32("i1");
        s->add_var(i32);
        delete i32; i32 = 0;
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "i1" && (*i++)->type() == dods_int32_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*v)->name() == "test.i1" && (*v++)->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }

    // with an array    
    void flatten_test2() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i1 = factory->NewInt32("i1");
        Array *a = factory->NewArray("i1", i1);
        delete i1; i1 = 0;
        s->add_var(a);
        delete a; a = 0;
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "i1" && (*i)->type() == dods_array_c
                       && (*i++)->var()->type() == dods_int32_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.i1" 
                       && (*v)->type() == dods_array_c
                       && (*v++)->var()->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }     
    
    // with a structure
    void flatten_test3() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i32 = factory->NewInt32("i1");
        Structure *ss = factory->NewStructure("ss");
        ss->add_var(i32);
        delete i32; i32 = 0;
        s->add_var(ss);
        delete ss; ss = 0;
        DBG2(cerr << "Before flatten: " << s->toString() << endl);
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "ss" && (*i++)->type() == dods_structure_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.ss.i1" && (*v++)->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }
    
    // with a grid
    void flatten_test4() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Grid *g = factory->NewGrid("g");
        Int32 *i1 = factory->NewInt32("i1");
        Array *a = factory->NewArray("a", i1);
        delete i1; i1 = 0;
        a->append_dim(100);
        Float64 *f1 = factory->NewFloat64("f1");
        Array *m = factory->NewArray("m", f1);
        delete f1; f1 = 0;
        a->append_dim(100);
        g->add_var(a, array);
        delete a; a = 0;
        g->add_var(m, maps);
        delete m; m = 0;
        s->add_var(g);
        delete g; g = 0;
        DBG2(cerr << "Before flatten: " << s->toString() << endl);
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "g" && (*i++)->type() == dods_grid_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.g" && (*v++)->type() == dods_grid_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);        
    }
    
    // with a sequence
    void flatten_test5() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i32 = factory->NewInt32("i1");
        Sequence *ss = factory->NewSequence("ss");
        ss->add_var(i32);
        delete i32; i32 = 0;
        s->add_var(ss);
        delete ss; ss = 0;
        DBG2(cerr << "Before flatten: " << s->toString() << endl);
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "ss" && (*i++)->type() == dods_sequence_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.ss.i1" && (*v)->type() == dods_array_c);
        CPPUNIT_ASSERT((*v++)->var()->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }

    // with an array of structures
    void flatten_test6() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i32 = factory->NewInt32("i1");
        Structure *ss = factory->NewStructure("a");
        ss->add_var(i32);
        delete i32; i32 = 0;
        Array *a = factory->NewArray("a");
        a->append_dim(200);
        a->add_var(ss);
        delete ss; ss = 0;
        s->add_var(a);
        delete a; a = 0;
        DBG2(cerr << "Before flatten: " << s->toString() << endl);
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "a" && (*i)->type() == dods_array_c);
        BaseType *btp = (*i)->var();
        CPPUNIT_ASSERT(btp->name() == "a" && btp->type() == dods_structure_c);
        CPPUNIT_ASSERT(++i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG2(cerr << "flattened var: " << (*v)->toString() << endl);
        CPPUNIT_ASSERT((*v)->name() == "test.a.i1" && (*v)->type() == dods_array_c);
        CPPUNIT_ASSERT((*v++)->var()->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }
    
    // with an array of sequences
    void flatten_test7() {
        Byte *b = factory->NewByte("b1");
        s->add_var(b);
        delete b; b = 0;
        Int32 *i32 = factory->NewInt32("i1");
        Sequence *ss = factory->NewSequence("as");
        ss->add_var(i32);
        delete i32; i32 = 0;
        Array *a = factory->NewArray("as", ss);
        delete ss; ss = 0;
        a->append_dim(200);
        s->add_var(a);
        delete a; a = 0;
        DBG(cerr << "Before flatten: "; s->print_decl(cerr, ""));
        Constructor::Vars_iter i = s->var_begin();
        CPPUNIT_ASSERT((*i)->name() == "b1" && (*i++)->type() == dods_byte_c);
        CPPUNIT_ASSERT((*i)->name() == "as" && (*i++)->type() == dods_array_c);
        CPPUNIT_ASSERT(i == s->var_end());
        
        VarList vars = s->flatten(*cp, "");
        VarListIter v = vars.begin();
        DBG(cerr << "flattened var: "; (*v)->print_decl(cerr, "") );
        CPPUNIT_ASSERT((*v)->name() == "test.b1" && (*v++)->type() == dods_byte_c);
        DBG(cerr << "flattened var: "; (*v)->print_decl(cerr, ""));
        CPPUNIT_ASSERT((*v)->name() == "test.as.i1" && (*v)->type() == dods_array_c);
        CPPUNIT_ASSERT((*v++)->var()->type() == dods_int32_c);
        CPPUNIT_ASSERT(v == vars.end());
        
        for_each(vars.begin(), vars.end(), delete_btp);
    }
    
};

CPPUNIT_TEST_SUITE_REGISTRATION(NCStructureTest);

int 
main( int argc, char* argv[] )
{
    CppUnit::TextTestRunner runner;
    runner.addTest( CppUnit::TestFactoryRegistry::getRegistry().makeTest() );

    bool wasSuccessful = runner.run( "", false ) ;

    return wasSuccessful ? 0 : 1;
}



